"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetResolvedInstructionAccountsVisitor = void 0;
const BaseThrowVisitor_1 = require("../BaseThrowVisitor");
class GetResolvedInstructionAccountsVisitor extends BaseThrowVisitor_1.BaseThrowVisitor {
    constructor() {
        super(...arguments);
        this.stack = [];
        this.visited = new Map();
        this.raw = new Map();
        this.error = null;
    }
    getError() {
        return this.error;
    }
    visitInstruction(instruction) {
        // Ensure we always start with a clean slate.
        this.error = null;
        this.stack = [];
        this.visited = new Map();
        this.raw = instruction.accounts.reduce((map, account, index) => map.set(account.name, [account, index]), new Map());
        // Visit all instruction accounts.
        this.raw.forEach(([account, index]) => {
            this.handleInstructionAccount(instruction, account, index);
        });
        return [...this.visited.values()].sort((a, b) => a.position - b.position);
    }
    handleInstructionAccount(instruction, account, index) {
        var _a, _b, _c;
        // Ensure we don't visit the same account twice.
        if (this.visited.has(account.name)) {
            return;
        }
        // Ensure we don't have a circular dependency.
        if (this.stack.includes(account.name)) {
            const cycle = [...this.stack, account.name].join(' -> ');
            this.error =
                `Circular dependency detected in the accounts of ` +
                    `the "${instruction.name}" instruction. ` +
                    `Got the following account dependency cycle: ${cycle}.`;
            throw new Error(this.error);
        }
        // Get account dependencies.
        const dependsOn = [];
        if (((_a = account.defaultsTo) === null || _a === void 0 ? void 0 : _a.kind) === 'account') {
            dependsOn.push(account.defaultsTo.name);
        }
        else if (((_b = account.defaultsTo) === null || _b === void 0 ? void 0 : _b.kind) === 'pda') {
            const accounts = Object.values(account.defaultsTo.seeds)
                .filter((seed) => seed.kind === 'account')
                .map((seed) => seed.name);
            dependsOn.push(...new Set(accounts));
        }
        // Visit account dependencies first.
        this.stack.push(account.name);
        dependsOn.forEach((name) => {
            const rawDependency = this.raw.get(name);
            if (!rawDependency) {
                this.error =
                    `Account "${name}" is not a valid dependency of account ` +
                        `"${account.name}" in the "${instruction.name}" instruction.`;
                throw new Error(this.error);
            }
            const [dependency, dependencyIndex] = rawDependency;
            this.handleInstructionAccount(instruction, dependency, dependencyIndex);
        });
        this.stack.pop();
        const resolved = {
            ...account,
            position: index,
            dependencyPosition: this.visited.size,
            dependsOn,
            resolvedIsSigner: account.isSigner,
            resolvedIsOptional: account.isOptional,
        };
        switch ((_c = resolved.defaultsTo) === null || _c === void 0 ? void 0 : _c.kind) {
            case 'account':
                const defaultAccount = this.visited.get(resolved.defaultsTo.name);
                const resolvedIsPublicKey = account.isSigner === false && defaultAccount.isSigner === false;
                const resolvedIsSigner = account.isSigner === true && defaultAccount.isSigner === true;
                const resolvedIsOptionalSigner = !resolvedIsPublicKey && !resolvedIsSigner;
                resolved.resolvedIsSigner = resolvedIsOptionalSigner
                    ? 'either'
                    : resolvedIsSigner;
                resolved.resolvedIsOptional = defaultAccount.isOptional;
                break;
            case 'publicKey':
            case 'program':
            case 'programId':
                resolved.resolvedIsSigner =
                    account.isSigner === false ? false : 'either';
                resolved.resolvedIsOptional = false;
                break;
            case 'pda':
                resolved.resolvedIsSigner =
                    account.isSigner === false ? false : 'either';
                resolved.resolvedIsOptional = false;
                const { seeds } = resolved.defaultsTo;
                Object.keys(seeds).forEach((seedKey) => {
                    const seed = seeds[seedKey];
                    if (seed.kind !== 'account')
                        return;
                    const dependency = this.visited.get(seed.name);
                    if (dependency.resolvedIsOptional) {
                        this.error =
                            `Cannot use optional account "${seed.name}" as the "${seedKey}" PDA seed ` +
                                `for the "${account.name}" account of the "${instruction.name}" instruction.`;
                        throw new Error(this.error);
                    }
                });
                break;
            case 'identity':
            case 'payer':
                resolved.resolvedIsOptional = false;
                break;
            default:
                break;
        }
        this.visited.set(account.name, resolved);
    }
}
exports.GetResolvedInstructionAccountsVisitor = GetResolvedInstructionAccountsVisitor;
//# sourceMappingURL=GetResolvedInstructionAccountsVisitor.js.map